from matplotlib.axes import Axes
from typing import Union, List, Optional, Tuple
import numpy as np
import shapely
from shapely.geometry import shape
from shapely import plotting
from geojson import GeoJSON, FeatureCollection
from ..objects.image_feature import ImageFeature


def plotImage(
    ax: Axes, image: np.ndarray, channel: int = 0, offset: Optional[Tuple] = None
) -> Axes:
    """
    A utility function to display a NumPy array returned by an ImageServer.

    :param ax: the Axes to plot to
    :param image: a numpy array with dimensions (c, y, x). If it represents a RGB image, all channels will be displayed.
                  Otherwise, only the specified channel will be displayed
    :param channel: the channel to display if the image is not RGB
    :param offset: an offset [x_min, y_min, x_max, y_max] to apply to the image. Useful if the image should match some features.
                   None to not define any offset
    :return: the Axes where the image was plotted
    """

    # Move channel axis
    image = np.moveaxis(image, [0], [2])

    # Keep only specified channel if the image is not RGB
    if image.shape[2] != 3 or image.dtype != "uint8":
        image = image[..., channel]

    # If pixel values are float numbers, they can be greater than 0, while matplotlib only accepts
    # values between 0 and 1, so the array is normalized between 0 and 1
    if "float" in str(image.dtype):
        image /= image.max()

    if offset is not None and len(offset) == 4:
        # y_min and y_max are swapped because otherwise image data is flipped vertically
        ax.imshow(image, extent=(offset[0], offset[2], offset[3], offset[1]))
    else:
        ax.imshow(image)

    return ax


def plotImageFeatures(
    ax: Axes,
    image_features: Union[FeatureCollection, List[ImageFeature]],
    region: Optional[GeoJSON] = None,
    default_color: tuple = (0.2, 0.8, 0.2),
    fill_opacity: float = 0.25,
) -> Axes:
    """
    A utility function to display a image features. This will plot the object's geometries
    and nucleus geometries.

    :param ax: the Axes to plot to
    :param image_features: the image features to plot
    :param region: the region to plot. Features outside this region won't be plotted.
                   All features are plotted if this parameter is undefined.
    :param default_color: the color to use for unclassified features
    :param fill_opacity: the opacity of the fill color. Set to 0 to not fill objects
    :return: the Axes where the image was plotted
    """
    envelope = None if region is None else shape(region).envelope
    if isinstance(image_features, FeatureCollection):
        image_features = image_features["features"]
    for feature in image_features:
        if feature.color is not None:
            edge_color = tuple(c / 255.0 for c in feature.color)
        elif feature.classification is not None:
            edge_color = tuple(c / 255.0 for c in feature.classification.color)
        else:
            edge_color = default_color

        if fill_opacity > 0:
            face_color = edge_color + (fill_opacity,)
        else:
            face_color = None

        for roi in [
            getattr(feature, name, None) for name in ["geometry", "nucleus_geometry"]
        ]:
            if roi is not None:
                geometry = shape(roi)
                if envelope is None or envelope.intersects(geometry.envelope):
                    _plot_geometry(ax, shape(roi), face_color, edge_color)

    ax.axis("equal")
    return ax


def _plot_geometry(
    ax: Axes, geometry: shapely.Geometry, face_color: Optional[Tuple], edge_color: Tuple
):
    if isinstance(geometry, Union[shapely.Point, shapely.MultiPoint]):
        plotting.plot_points(geometry, ax, color=edge_color)
    elif isinstance(geometry, Union[shapely.LineString, shapely.LinearRing]):
        plotting.plot_line(geometry, ax, add_points=False, color=edge_color)
    elif isinstance(geometry, Union[shapely.Polygon, shapely.MultiPolygon]):
        plotting.plot_polygon(
            geometry,
            ax,
            add_points=False,
            facecolor=face_color,
            edgecolor=edge_color,
            linewidth=2,
        )
    elif isinstance(geometry, shapely.MultiLineString):
        for line in geometry.geoms:
            plotting.plot_line(line, ax, add_points=False, color=edge_color)
    elif isinstance(geometry, shapely.GeometryCollection):
        for g in geometry.geoms:
            _plot_geometry(ax, g, face_color, edge_color)



import geojson.geometry
import geojson
from typing import Union, Optional


class ImagePlane(dict):
    """
    An additional property of a geometry containing z and t indices.
    """

    def __init__(self, z: int = 0, t: int = 0):
        """
        :param z: the z-stack of the geometry
        :param t: the time point of the geometry
        """
        dict.__init__(self, z=z, t=t)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        raise AttributeError("ImagePlane is immutable")

    def __delattr__(self, name):
        raise AttributeError("ImagePlane is immutable")


def add_plane_to_geometry(
    geometry: Union[geojson.geometry.Geometry, geojson.Feature],
    z: Optional[int] = None,
    t: Optional[int] = None,
    preferred_geometry_key: Optional[str] = None,
) -> Optional[geojson.geometry.Geometry]:
    """
    Create a GeoJSON Geometry object with an additional 'plane' property containing z and t indices.

    If indices are unspecified, they will be taken from the geometry's 'plane' property, if present,
    or else z or t attributes otherwise.
    If z and t values cannot be found in either of these locations, defaults (z=0 and t=0) will be
    used instead.

    The 'plane' property is immutable.

    :param geometry: a GeoJSON feature or geometry. It must not be a FeatureCollection
    :param z: the z-stack index this geometry should have
    :param t: the timepoint this geometry should have
    :param preferred_geometry_key: if the provided geometry is a feature, and this feature has a 'extra_geometries'
                                   property, then this parameter can be used as a key to retrieve the geometry to
                                   consider from the 'extra_geometries' dictionnary
    :return: a GeoJSON Geometry containing an additional 'plane' property containing z and t indices
    :raises ValueError: when the provided geometry is a FeatureCollection
    """
    if geometry is None:
        return None

    if isinstance(geometry, geojson.Feature):
        feature = geometry
        if (
            preferred_geometry_key is not None
            and "extra_geometries" in feature.properties
        ):
            geometry = feature.properties["extra_geometries"].get(
                preferred_geometry_key, feature.geometry
            )
        else:
            geometry = feature.geometry
    elif isinstance(geometry, geojson.FeatureCollection):
        raise ValueError("Cannot convert FeatureCollection to single Geometry")

    if z is None or t is None:
        plane = getattr(geometry, "plane", None)
        if z is None:
            z = (
                getattr(plane, "z", 0)
                if plane is not None
                else getattr(geometry, "z", 0)
            )
        if t is None:
            t = (
                getattr(plane, "t", 0)
                if plane is not None
                else getattr(geometry, "t", 0)
            )

    geometry = geojson.GeoJSON.to_instance(geometry, strict=False)
    geometry.plane = ImagePlane(z, t)
    return geometry



from .qupath_gateway import (
    get_default_gateway,
    set_default_gateway,
    create_gateway,
    count_objects,
    create_snapshot,
    get_current_image_data,
    add_objects,
    delete_objects,
    get_objects,
    get_project,
    refresh_qupath,
)



from __future__ import annotations
import geojson
import uuid
import math
import numpy as np
import geojson.geometry
from typing import Union, Optional, Any
import rasterio
import rasterio.features
import shapely
from .object_type import ObjectType
from .classification import Classification
from .geometry import add_plane_to_geometry


_NUCLEUS_GEOMETRY_KEY = "nucleus"


class ImageFeature(geojson.Feature):
    """
    GeoJSON Feature with additional properties for image objects.

    The added properties are:

    - A classification (defined by a name and a color).

    - A name.

    - A list of measurements.

    - A type of QuPath object (e.g. detection, annotation).

    - A color.

    - Additional geometries.

    - And any other property.
    """

    def __init__(
        self,
        geometry: Optional[geojson.geometry.Geometry],
        classification: Optional[Union[Classification, dict]] = None,
        name: Optional[str] = None,
        measurements: Optional[dict[str, float]] = None,
        object_type: ObjectType = ObjectType.ANNOTATION,
        color: Optional[tuple[int, int, int]] = None,
        extra_geometries: Optional[dict[str, geojson.geometry.Geometry]] = None,
        id: Optional[Union[str, int, uuid.UUID]] = None,
        extra_properties: Optional[dict[str, Any]] = None,
    ):
        """
        Except from the geometry and id parameters, all parameters of this
        constructor will be added to the list of properties of this feature
        (if provided).

        :param geometry: the geometry of the feature
        :param classification: the classification of this feature, or a dictionnary with the
                               'name' and 'color' properties defining respectively a string
                               and a 3-long int tuple with values between 0 and 255
        :param name: the name of this feature
        :param measurements: a dictionnary containing measurements. Measurements
                             with NaN values will not be added
        :param object_type: the type of QuPath object this feature represents
        :param color: the color of this feature
        :param extra_geometries: a dictionnary containing additional geometries
                                 that represent this feature
        :param id: the ID of the feature. If not provided, an UUID will be generated
        :param extra_properties: a dictionnary of additional properties to add
        """
        props = {}
        if classification is not None:
            if isinstance(classification, Classification):
                props["classification"] = {
                    "names": classification.names,
                    "color": classification.color,
                }
            else:
                props["classification"] = {
                    "names": classification.get("names")
                    if "names" in classification.keys()
                    else (classification.get("name"),),
                    "color": classification.get("color"),
                }
        if name is not None:
            props["name"] = name
        if measurements is not None:
            props["measurements"] = ImageFeature._remove_NaN_values_from_measurements(
                measurements
            )
        if object_type is not None:
            props["object_type"] = object_type.name
        if color is not None:
            props["color"] = color
        if extra_geometries is not None:
            props["extra_geometries"] = {
                k: add_plane_to_geometry(v) for k, v in extra_geometries.items()
            }
        if extra_properties is not None:
            props.update(extra_properties)

        super().__init__(
            geometry=add_plane_to_geometry(geometry),
            properties=props,
            id=ImageFeature._to_id_string(id),
        )
        self["type"] = "Feature"

    @classmethod
    def create_from_feature(cls, feature: geojson.Feature) -> ImageFeature:
        """
        Create an ImageFeature from a GeoJSON feature.

        The ImageFeature properties will be searched in the provided
        feature and in the properties of the provided feature.

        :param feature: the feature to convert to an ImageFeature
        :return: an ImageFeature corresponding to the provided feature
        """
        geometry = cls._find_property(feature, "geometry")
        plane = cls._find_property(feature, "plane")
        if plane is not None:
            geometry = add_plane_to_geometry(
                geometry, z=getattr(plane, "z", None), t=getattr(plane, "t", None)
            )

        object_type_property = cls._find_property(feature, "object_type")
        if object_type_property is None:
            object_type_property = cls._find_property(feature, "objectType")
        object_type = next(
            (
                o
                for o in ObjectType
                if o.name.lower() == str(object_type_property).lower()
            ),
            None,
        )

        args = dict(
            geometry=geometry,
            id=cls._find_property(feature, "id"),
            classification=cls._find_property(feature, "classification"),
            name=cls._find_property(feature, "name"),
            color=cls._find_property(feature, "color"),
            measurements=cls._find_property(feature, "measurements"),
            object_type=object_type,
        )

        nucleus_geometry = cls._find_property(feature, "nucleusGeometry")
        if nucleus_geometry is not None:
            if plane is not None:
                nucleus_geometry = add_plane_to_geometry(
                    nucleus_geometry,
                    z=getattr(plane, "z", None),
                    t=getattr(plane, "t", None),
                )
            args["extra_geometries"] = dict(nucleus=nucleus_geometry)

        args["extra_properties"] = {
            k: v
            for k, v in feature["properties"].items()
            if k not in args and v is not None
        }
        return cls(**args)

    @classmethod
    def create_from_label_image(
        cls,
        input_image: np.ndarray,
        object_type: ObjectType = ObjectType.ANNOTATION,
        connectivity: int = 4,
        scale: float = 1.0,
        include_labels=False,
        classification_names: Optional[Union[str, dict[int, str]]] = None,
    ) -> list[ImageFeature]:
        """
        Create a list of ImageFeatures from a binary or labeled image.

        The created geometries will be polygons, even when representing points or line.

        :param input_image: a 2-dimensionnal binary (with a boolean type) or labeled
                            (with a uint8 type) image containing the features to create.
                            If a binary image is given, all True pixel values will be
                            considered as potential features. If a labeled image is given,
                            all pixel values greater than 0 will be considered as potential features
        :param object_type: the type of object to create
        :param connectivity: the pixel connectivity for grouping pixels into features (4 or 8)
        :param scale: a scale value to apply to the shapes
        :param include_labels: whether to include a 'Label' measurement in the created features
        :param classification_names: if str, the name of the classification to apply to all features.
                                     if dict, a dictionnary mapping a label to a classification name
        :return: a list of image features representing polygons present in the input image
        """
        features = []

        if input_image.dtype == bool:
            mask = input_image
            input_image = input_image.astype(np.uint8)
        else:
            mask = input_image > 0

        transform = rasterio.transform.Affine.scale(scale)

        existing_features = {}
        for geometry, label in rasterio.features.shapes(
            input_image, mask=mask, connectivity=connectivity, transform=transform
        ):
            if label in existing_features:
                existing_features[label]["geometry"] = shapely.geometry.shape(
                    geometry
                ).union(shapely.geometry.shape(existing_features[label]["geometry"]))
            else:
                if isinstance(classification_names, str):
                    classification_name = classification_names
                elif (
                    isinstance(classification_names, dict)
                    and int(label) in classification_names
                ):
                    classification_name = classification_names[int(label)]
                else:
                    classification_name = None

                feature = cls(
                    geometry=geometry,
                    classification=Classification(classification_name),
                    measurements={"Label": float(label)} if include_labels else None,
                    object_type=object_type,
                )

                existing_features[label] = feature
                features.append(feature)

        # Ensure we have GeoJSON-compatible geometries
        for feature in features:
            feature["geometry"] = geojson.mapping.to_mapping(feature["geometry"])

        return features

    @property
    def classification(self) -> Classification:
        """
        The classification of this feature (or None if not defined).
        """
        if "classification" in self.properties:
            return Classification(
                self.properties["classification"].get("names")
                if "names" in self.properties["classification"].keys()
                else (self.properties["classification"].get("name"),),
                self.properties["classification"].get("color"),
            )
        else:
            return None

    @property
    def name(self) -> str:
        """
        The name of this feature (or None if not defined).
        """
        return self.properties.get("name")

    @property
    def measurements(self) -> dict[str, float]:
        """
        The measurements of this feature.
        """
        measurements = self.properties.get("measurements")
        if measurements is None:
            measurements = {}
            self.properties["measurements"] = measurements
        return measurements

    @property
    def object_type(self) -> ObjectType:
        """
        The QuPath object type (e.g. detection, annotation) this feature represents
        or None if the object type doesn't exist or is not recognised.
        """
        return next(
            (
                o
                for o in ObjectType
                if o.name.lower() == str(self.properties["object_type"]).lower()
            ),
            None,
        )

    @property
    def is_detection(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a detection, cell, or tile.
        """
        return self.object_type in [
            ObjectType.DETECTION,
            ObjectType.CELL,
            ObjectType.TILE,
        ]

    @property
    def is_cell(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a cell.
        """
        return self.object_type == ObjectType.CELL

    @property
    def is_tile(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is a tile.
        """
        return self.object_type == ObjectType.TILE

    @property
    def is_annotation(self) -> bool:
        """
        Wether the QuPath object type (e.g. detection, annotation) represented by this
        feature is an annotation.
        """
        return self.object_type == ObjectType.ANNOTATION

    @property
    def color(self) -> tuple[int, int, int]:
        """
        The color of this feature (or None if not defined).
        """
        return self.properties.get("color")

    @property
    def nucleus_geometry(self) -> geojson.geometry.Geometry:
        """
        The nucleus geometry of this feature (or None if not defined).
        It can be defined when passed as an extra_geometry with the 'nucleus'
        key when creating an ImageFeature, by defining the 'nucleus_geometry'
        property of an ImageFeature, or when passed as a 'nucleusGeometry'
        property when creating an ImageFeature from a GeoJSON feature.
        """
        extra = self.properties.get("extra_geometries")
        if extra is not None:
            return extra.get(_NUCLEUS_GEOMETRY_KEY)
        return None

    def __setattr__(self, name, value):
        if name == "classification":
            if isinstance(value, Classification):
                self.properties["classification"] = {
                    "name": value.name,
                    "color": value.color,
                }
            else:
                self.properties["classification"] = value
        elif name == "name":
            self.properties["name"] = value
        elif name == "measurements":
            self.properties[
                "measurements"
            ] = ImageFeature._remove_NaN_values_from_measurements(value)
        elif name == "object_type":
            if isinstance(value, str):
                self.properties["object_type"] = value
            elif isinstance(value, ObjectType):
                self.properties["object_type"] = value.name
        elif name == "color":
            if len(value) != 3:
                raise ValueError("Color must be a tuple of length 3")
            rgb = tuple(ImageFeature._validate_rgb_value(v) for v in value)
            self.properties["color"] = rgb
        elif name == "nucleus_geometry":
            if "extra_geometries" not in self.properties:
                self.properties["extra_geometries"] = {}
            self.properties["extra_geometries"][
                _NUCLEUS_GEOMETRY_KEY
            ] = add_plane_to_geometry(value)
        else:
            super().__setattr__(name, value)

    @staticmethod
    def _remove_NaN_values_from_measurements(
        measurements: dict[str, float]
    ) -> dict[str, float]:
        return {
            k: float(v)
            for k, v in measurements.items()
            if isinstance(k, str) and isinstance(v, (int, float)) and not math.isnan(v)
        }

    @staticmethod
    def _to_id_string(object_id: Optional[Union[int, str, uuid.UUID]]) -> str:
        if object_id is None:
            return str(uuid.uuid4())
        else:
            return str(object_id)

    @staticmethod
    def _find_property(feature: geojson.Feature, property_name: str):
        if property_name in feature:
            return feature[property_name]
        elif "properties" in feature and property_name in feature["properties"]:
            return feature["properties"][property_name]
        else:
            return None

    @staticmethod
    def _validate_rgb_value(value: Union[int, float]) -> int:
        if isinstance(value, float):
            value = int(round(value * 255))
        if isinstance(value, int):
            if value >= 0 and value <= 255:
                return value
        raise ValueError(
            "Color value must be an int between 0 and 255, or a float between 0 and 1"
        )



from .bioio_server import BioIOServer
from .icc_profile_server import IccProfileServer
from .image_server import ImageServer
from .labeled_server import LabeledImageServer
from .openslide_server import OpenSlideServer
from .qupath_server import QuPathServer
from .region_2d import Region2D
from .utils import base64_to_image, bytes_to_image
from .wrapped_image_server import WrappedImageServer



from __future__ import annotations
import random
from typing import Optional, Union


class Classification(object):
    """
    Simple class to store the names and color of a classification.

    Each Classification with the same names is the same object, retrieved from a cache.
    Therefore updating the color of a Classification will update all similarly classified objects.
    """

    _cached_classifications: dict[tuple[str], Classification] = {}

    def __new__(
        cls, names: Union[str, tuple[str]], color: Optional[tuple[int, int, int]] = None
    ):
        if isinstance(names, str):
            names = (names,)
        elif isinstance(names, list):
            names = tuple(names)
        if names is None:
            return None
        if not isinstance(names, tuple):
            raise TypeError("names should be str or tuple[str]")

        classification = Classification._cached_classifications.get(names)
        if classification is None:
            classification = super().__new__(cls)
            Classification._cached_classifications[names] = classification

        if color is not None:
            classification.color = color
        return classification

    def __init__(
        self,
        names: Union[str, tuple[str]],
        color: Optional[tuple[int, int, int]] = None,
    ):
        """
        :param names: the names of the classification
        :param color: the RGB color (each component between 0 and 255) of the classification. Can be None to use a random color
        """
        if isinstance(names, str):
            names = (names,)
        elif isinstance(names, list):
            names = tuple(names)
        if not isinstance(names, tuple):
            raise TypeError("names should be a tuple, list or string")
        self._names = names
        self._color: tuple = (
            tuple(random.randint(0, 255) for _ in range(3)) if color is None else color
        )

    @property
    def names(self) -> tuple[str]:
        """
        The names of the classification.
        """
        return self._names

    @property
    def color(self) -> tuple[int, int, int]:
        """
        The color of the classification.
        """
        return self._color  ## todo: pylance type hints problem

    @color.setter
    def color(self, value: tuple[int, int, int]) -> None:
        """
        Change the color of the classification.
        :param value: the new 8-bit RGB color
        """
        self._color = value

    def __str__(self):
        return f"Classification {self.names} of color {self.color}"

    def __repr__(self):
        return f"Classification('{self.names}', {self.color})"

    def __eq__(self, other):
        if isinstance(other, Classification):
            return (self is other) or (self.names == other.names)
        return False

    def __hash__(self):
        return hash(self.names)



from .object_type import ObjectType
from .classification import Classification
from .geojson import geojson_features_from_string
from .draw import draw_geometry
from .image_feature import ImageFeature
from .geometry import ImagePlane, add_plane_to_geometry
